#include "../../std.h"
#include <iostream>

#include "../../misc/thread.h"
#include "../../misc/log_time.h"
#include "file.h"
#include "se_ns.h"


#undef SOAP_FMAC1
#undef SOAP_FMAC3
#undef SOAP_FMAC5
#undef SOAP_NMAC
#define WITH_NONAMESPACES
#define WITH_NOGLOBAL
#include "stdsoap2.h"
#undef SOAP_FMAC1
#undef SOAP_FMAC3
#undef SOAP_FMAC5
#undef SOAP_NMAC
#define SOAP_FMAC1 static
#define SOAP_FMAC3 static
#define SOAP_FMAC5 static
#define SOAP_NMAC static
#include "ns.nsmap"
#include "file_soapC.cpp"
#include "file_soapServer.cpp"

#define __SE_INTERNAL_INCLUDE__
#include "../se.h"

ns__fileinfo::ns__fileinfo() {
}
ns__fileinfo::~ns__fileinfo() {
}

static bool merge_urls(string &cfg_url,const char* contact_base) {
  if((cfg_url.length() == 0) || (cfg_url[0] == '/')) { // path
    const char* u_ = contact_base;
    u_=strchr(u_,':'); if(u_==NULL) return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; u_=strchr(u_,'/'); if(u_==NULL) u_=contact_base+strlen(contact_base);
    string url(contact_base,u_-contact_base);
    cfg_url=url+cfg_url;
  } else if(cfg_url[0] == ':') { // port+path
    const char* u_ = contact_base;
    u_=strchr(u_,':'); if(u_==NULL) return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; u_=strchr(u_,'/'); if(u_==NULL) u_=contact_base+strlen(contact_base);
    int p = cfg_url.find('/'); if(p == string::npos) p=cfg_url.length();
    string url(contact_base,u_-contact_base);
    url+=cfg_url.c_str()+p;
    cfg_url=url;
  } else { // url

  };
  return true;
}

HTTP_Service* se_service_creator(HTTPS_Connector& c,const char* uri,void* arg) {
  HTTP_SE* h = new HTTP_SE(&c,(HTTP_SE_Handle*)arg,uri);
  return h;
}

#define registration_immediate 1
#define registration_retry 2
class HTTP_SE_Handle {
 friend class HTTP_SE;
 friend bool se_service_configurator(istream& f,const char* uri,HTTP_Service_Properties &prop);
 private:
  // URL used to announce available files
  string base_url;
  // files handled by this service
  SEFiles files;
  int reg_type;
 public:
  HTTP_SE_Handle(const char* dirpath,const char* url):files(dirpath),base_url(url) { };
  ~HTTP_SE_Handle(void) { };
  operator bool(void) { return (bool)(files); };
  bool register_immediately(void) {
    return (reg_type & registration_immediate)?true:false;
  };
  bool register_retry(void) {
    return (reg_type & registration_retry)?true:false;
  };
  SENameServer* NS(void) { return files.NS(); };
};

class HTTP_SE_Thread: public Thread {
 private:
  LockSimple lock;
  list<SEFiles*> files;
  virtual void func(void);
 public:
  HTTP_SE_Thread(void);
  ~HTTP_SE_Thread(void);
  void add_files(SEFiles* f);
};

HTTP_SE_Thread::HTTP_SE_Thread(void) {



}

HTTP_SE_Thread::~HTTP_SE_Thread(void) {


}

void HTTP_SE_Thread::add_files(SEFiles* f) {
  lock.block();
  files.push_back(f);
  lock.unblock();
}

void HTTP_SE_Thread::func(void) {
  while(true) {
    lock.block();
    // go through all storages
cerr<<"HTTP_SE_Thread::func: awake"<<endl;
    for(list<SEFiles*>::iterator fs = files.begin();fs!=files.end();++fs) {
      // go through all files
      SEFiles* fsp = (*fs);
      // check if it is time to retry registration (TODO)
cerr<<"HTTP_SE_Thread::func: Register"<<endl;
      if(*fs) (*fs)->Register();
    };
    lock.unblock();
cerr<<"HTTP_SE_Thread::func: sleep"<<endl;
    idle(10*60*1000); // 10 minutes
  };
}

static HTTP_SE_Thread* se_thread = NULL;

HTTP_SE::HTTP_SE(HTTPS_Connector *c_,HTTP_SE_Handle *h_,const char* uri_):HTTP_ServiceAdv(c_),handle(h_),url(uri_) {
  odlog(0)<<"SE: new service: service URL: "<<uri_<<endl;
  odlog(0)<<"SE: new service: contacted URL: "<<c_->url()<<endl;
  odlog(0)<<"SE: new service: configured URL: "<<h_->base_url<<endl;
  if(h_->base_url.length()) url=h_->base_url;
  merge_urls(url,c_->url());
  odlog(0)<<"SE: new service: files URL: "<<url<<endl;
  soap_init();
  sp.namespaces=namespaces;
  sp.user=this;
}

HTTP_SE::~HTTP_SE(void) {
  soap_deinit();
}

SEFiles* HTTP_SE::files(void) {
  if(handle==NULL) return NULL;
  return &(handle->files);
}

/*
 * HTTP methods
 */
HTTP_Error HTTP_SE::get(const char* uri,int &keep_alive) {
  if((!c) || (!handle)) {
    return HTTP_NOT_IMPLEMENTED;
  };
  HTTP_Error err;
  if((err=parse_header(keep_alive)) != HTTP_OK) return err;
  if(failure_parsing || unsupported_option_passed) {
    keep_alive=0;
    return HTTP_FAILURE;
  };
  // obtain requested name
  const char* name = uri+handle->base_url.length();
  if(*name != 0) name++; // skip slash
  // get file
  SEFileHandle* h = SEFileHandle::open(name,true,handle->files);
  if(h==NULL) {
cerr<<"file not found"<<endl;
    if(send_error_response(*c,keep_alive,404,NULL,"404 Not Found") != 0) {
      keep_alive=0;
      return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  // fix ranges
  uint64_t range_start_,range_end_;
  if(range_passed) {
cerr<<"range_passed"<<endl;
    int n;  
    for(n=0;n<nranges;n++) { // find first suitable range
      if(range_start[n] <= range_end[n]) {
        if(range_start[n] < h->size()) {
          range_end[n]++;
          if(range_end[n] > h->size()) range_end[n]=h->size();
          break;
        };
      } else {
        if(range_start[n] > h->size()) range_start[n]=h->size();
        range_start[n]=h->size()-range_start[n];
        range_end[n]=h->size();
        break;
      };
    };
    if(n<nranges) { range_start_=range_start[n]; range_end_=range_end[n]; }
    else { range_start_=0; range_end_=0; }; // no such data
    h->seek(range_start_);
  } else {
cerr<<"range_not_passed"<<endl;
    range_start_=0; range_end_=h->size();
  };
cerr<<"range_start: "<<range_start_<<endl;
cerr<<"range_end: "<<range_end_<<endl;
cerr<<"size: "<<h->size()<<endl;
  // send reply header
  if(range_end_ <= range_start_) {
cerr<<"range_end <= range_start"<<endl;
    delete h;
    if(send_error_response(*c,keep_alive,416,NULL,"Error 416") != 0) {
cerr<<"failed to send 416"<<endl;
      keep_alive=0;
      return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  bool partial=false;
  if((range_start_ != 0) || (range_end_ != h->size())) partial=true;
cerr<<"partial: "<<partial<<endl;
  err=send_header(keep_alive,range_start_,range_end_,partial,h->size());
cerr<<"send_header: "<<err<<endl;
  if(err != HTTP_OK) { keep_alive=0; delete h; return HTTP_ERROR; };
  // pass requested range to client
  err=HTTP_OK;
  for(;;) {
    if(h->tell() >= range_end_) break; 
    uint64_t l = range_end_-h->tell();
    char buf[65536];
    if(l > sizeof(buf)) l=sizeof(buf);
    l=h->read(buf,l);
    if(l==0) break; // eof or error
    if(c->write(buf,l) != 0) {
      cerr<<"HTTP_SE::get: failure while reading data from disc"<<endl;
      keep_alive=0;
      err=HTTP_FAILURE;
    };
  };
  if(err==HTTP_OK) if(h->tell() < range_end_) { // something wrong with file
    cerr<<"HTTP_SE::get: failure while reading data from disc"<<endl;
    keep_alive=0;
    err=HTTP_FAILURE;
  };
  delete h;
cerr<<"HTTP_SE: normal exit: return"<<endl;
  return err;
}

HTTP_Error HTTP_SE::put(const char* uri,int &keep_alive) {
  if((!c) || (!handle)) {
    return HTTP_NOT_IMPLEMENTED;
  };
  HTTP_Error err;
  if((err=parse_header(keep_alive)) != HTTP_OK) return err;
  if(failure_parsing || unsupported_option_passed) {
    cerr<<"HTTP_SE::put: bad header"<<endl;
    keep_alive=0;
    if(send_header(keep_alive,406) != HTTP_OK) keep_alive=0;
    return HTTP_FAILURE;
  };
  // obtain requested name
  const char* name = uri+handle->base_url.length();
  if(*name != 0) name++; // skip slash
  cerr<<"HTTP_SE::put: name: "<<name<<endl;
  // check state (not yet!)



  // open file
  SEFileHandle* h = SEFileHandle::open(name,false,handle->files);
  if(h==NULL) {
    cerr<<"HTTP_SE::put: file not found: "<<name<<endl;
    if(send_error_response(*c,keep_alive,404,NULL,"404 Not Found") != 0) {
      keep_alive=0;
      return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  uint64_t range_start_,range_end_;
  if(entity_range_passed) {
    cerr<<"HTTP_SE::put: range_passed: "<<range_start_<<"-"<<range_end_<<endl;
    range_start_=entity_range_start;
    range_end_=entity_range_end+1;
    if(length_passed && ((range_end_-range_start_) != length)) { 
      keep_alive=0; delete h;
      if(send_header(keep_alive,406) != HTTP_OK) keep_alive=0;
      return HTTP_FAILURE;
    };
  } else {
    cerr<<"HTTP_SE::put: range_not_passed"<<endl;
    range_start_=0; range_end_=h->size();
    if(length_passed) range_end_=length;
  };
  if(range_end_ > h->size()) { 
    keep_alive=0; delete h;
    if(send_header(keep_alive,500) != HTTP_OK) keep_alive=0;
    return HTTP_FAILURE;
  };
  h->seek(range_start_);
  err=HTTP_OK;
  // receive entity from network and store locally 
  for(;;) {
    if(h->tell() >= range_end_) break; 
    uint64_t l = range_end_-h->tell();
    char buf[65536];
    if(l > sizeof(buf)) l=sizeof(buf);
    l=c->read(buf,l);
    if(l == 0) break; // EOF
    l=h->write(buf,l);
    if(l==0) { // error
      odlog(0)<<"Failed to write to file"<<endl;
      err=HTTP_FAILURE;
      break;
    };
  };
  if((err!=HTTP_OK) || (h->tell() < range_end_)) { // something wrong with file
    cerr<<"HTTP_SE::put: failure while reading/writing data from/to network/disc"<<endl;
    keep_alive=0;
    if(send_header(keep_alive,500) != HTTP_OK) keep_alive=0;
    err=HTTP_FAILURE;
  };
  h->close();
  if(h->full()) {                                   // full content stored
    SEFile& f = h->instance();
    if((f.complete()) &&                            // metadata complete
       (f.state_reg() == REG_STATE_LOCAL) &&        // not registered yet
       (!(handle->register_immediately()))) {       // register after transfer
          // register new file
      if(f.state(FILE_STATE_UNCHANGED,REG_STATE_REGISTERING)) {
        if(handle->NS()->Register(f) != 0) {
          if(handle->register_retry()) {
            cerr<<"Failed to register (will retry)"<<endl;
          } else {
            cerr<<"Failed to register (what todo ?)"<<endl;
          };
          f.state(FILE_STATE_UNCHANGED,REG_STATE_LOCAL);
        } else {
          f.state(FILE_STATE_UNCHANGED,REG_STATE_ANNOUNCED);
        };
      } else {
        if(handle->register_retry()) {
          cerr<<"Failed to set REGISTERING state (will retry)"<<endl;
        } else {
          cerr<<"Failed to set REGISTERING state (waht to do ?)"<<endl;
        };
      };
    };
    f.state(FILE_STATE_COMPLETE,REG_STATE_UNCHANGED);
  };
  delete h;
  if(err==HTTP_OK) if(send_header(keep_alive,200) != HTTP_OK) keep_alive=0;
  return err;
}

HTTP_Error HTTP_SE::post(const char* uri,int &keep_alive) {
  // obtain requested name
cerr<<"post: uri: "<<uri<<endl;
  current_name = uri+handle->base_url.length();
  // get file
  current_file = handle->files.get(current_name);
  if(current_file==NULL) {
/*
    skip_header(*c,keep_alive);
    if(send_error_response(*c,keep_alive,404,NULL,"404 Not Found") != 0) {
      keep_alive=0;
      return HTTP_ERROR;
    };
    return HTTP_OK;
*/
  };
  enum HTTP_Error r = soap_post(uri,keep_alive);
  if(current_file) { delete current_file; current_file=NULL; };
  return r;
}

// called during startup of server (once for every base uri)
bool se_service_configurator(istream& f,const char* uri,HTTP_Service_Properties &prop) {
  string path("");
  string ns_url("");
  string base_url("");
  int reg_type = registration_immediate;
  for(;!f.eof();) {
    char buf[1024];
    istream_readline(f,buf,sizeof(buf));
    char* p = buf;
    for(;*p;p++) if(!isspace(*p)) break;
    char* command = p;
    if(*command == '#') continue;
    for(;*p;p++) if(isspace(*p)) break;
    int command_len = p-command;
    if(!command_len) continue;
    if((command_len==4) && (strncmp(command,"path",command_len)==0)) {
      for(;*p;p++) if(!isspace(*p)) break;
      path=p;
      continue;
    } else if((command_len==2) && (strncmp(command,"ns",command_len)==0)) {
      for(;*p;p++) if(!isspace(*p)) break;
      ns_url=p;
      continue;
    } else if((command_len==3) && (strncmp(command,"url",command_len)==0)) {
      for(;*p;p++) if(!isspace(*p)) break;
      base_url=p;
      continue;
    } else if((command_len==12) && (strncmp(command,"registration",command_len)==0)) {
      // parse options separated by ,
      for(;;) {
        for(;*p;p++) if(!isspace(*p)) break;
        if(!(*p)) break;
        char* pc = p;
        for(;*pc;pc++) if((*p)==',') break;
        if(strncmp(p,"immediate",pc-p) == 0) { reg_type|=registration_immediate; }
        else if(strncmp(p,"delayed",pc-p) == 0) { reg_type&=~((unsigned int)(registration_immediate)); }
        else if(strncmp(p,"retry",pc-p) == 0) { reg_type|=registration_retry; }
        else if(strncmp(p,"noretry",pc-p) == 0) { reg_type&=~((unsigned int)(registration_retry)); };
        p=pc; if(*p) p++;
      };
    } else if((command_len==3) && (strncmp(command,"end",command_len)==0)) {
      break;
    } else {
      odlog(0)<<"SE(configure): skipping unknow command: "<<command<<endl;
    };
  };
  prop.arg=NULL;
  if(base_url.length() == 0) base_url=uri;
  odlog(0)<<"SE(configure): filesystem path: "<<path<<endl;
  odlog(0)<<"SE(configure): URL: "<<base_url<<endl;
  HTTP_SE_Handle* handle = NULL;
  if(path.length()) {
    handle = new HTTP_SE_Handle(path.c_str(),base_url.c_str());
    handle->reg_type=reg_type;
    if((*handle)) { prop.arg=handle; }
    else { delete handle; handle=NULL; };
  };
  if(!handle) {
    odlog(0)<<"SE(configure): failed"<<endl;
  };
  if(handle) {
    handle->files.NS(ns_url.c_str(),uri);
  };
  prop.subtree=true;
  odlog(0)<<"SE(configure): serving subtree"<<endl;
  // run thread to handle delayed registration and other things
  // one thread per server
  if(se_thread == NULL) {
    se_thread = new HTTP_SE_Thread;
  };
  if(handle) {
    se_thread->add_files(&(handle->files));
  };
  return true;
}

SEFile* HTTP_SE::new_file(SEAttributes& attr) {
  if((!c) || (!handle)) {
    return NULL;
  };
  odlog(0)<<"SE: creating new file: "<<attr.id()<<endl;
  SEFile* f = new SEFile(handle->files.Path(),attr);
  if(f) { if(!(*f)) { delete f; f=NULL; }; };
  if(!f) { odlog(0)<<"SE: file creation failed"<<endl; return NULL; };
  return handle->files.add(*f);
}

void HTTP_SE::soap_methods(void) {
  if((sp.error = soap_serve_ns__add(&sp)) != SOAP_NO_METHOD) return;
  if((sp.error = soap_serve_ns__update(&sp)) != SOAP_NO_METHOD) return;
  if((sp.error = soap_serve_ns__info(&sp)) != SOAP_NO_METHOD) return;
  if((sp.error = soap_serve_ns__acl(&sp)) != SOAP_NO_METHOD) return;
}

char* soap_strdup_l(soap* sp,const char* s,int l = -1) {
  if(l<0) l=strlen(s);
  char* s_ = (char*)soap_malloc(sp,l+1);
  if(s_ == NULL) return NULL;
  memcpy(s_,s,l); s_[l]=0;
  return s_;
}

void add_path_to_url(string& url,const string& path,bool strip_path = true) {
  int l = url.length()-1;
  if(l == 0) { url=path; return; };
  if(url[l] != '/') url+="/";
  const char* s = path.c_str();
  if(strip_path) if(s[0] == '/') s++;
  url+=s;
  return;
}

/* SOAP calls */
int ns__update(struct soap *sp, ns__fileinfo *file, struct ns__updateResponse &r) {
  // initialize response
  r.error_code=0; r.sub_error_code=0;
  r.file.id=NULL;
  r.file.size=NULL;
  r.file.checksum=NULL;
  r.file.acl=NULL;
  r.file.url=NULL;
  r.file.state=NULL;
  //
  // update file attributes and check if state should not be changed 
  //
  if(!file) {
    cerr<<"SOAP: update: missing information about file"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA;
    return SOAP_OK;
  };
  if(file->url) {
    cerr<<"SOAP: update: update does not need contact url"<<endl;
    r.error_code=ERR_EXCESSIVE_INPUT_DATA;
    return SOAP_OK;
  };
  HTTP_SE* it = (HTTP_SE*)(sp->user);
  if(file->id && it->current_name && (*it->current_name) && (strcmp(it->current_name,file->id)!=0)) {
    cerr<<"SOAP: update: if provided file id must be same as path"<<endl;
    r.error_code=ERR_WRONG_INPUT_DATA;
    return SOAP_OK;
  };
  if((!(file->id)) && (it->current_name || (*it->current_name))) {
    cerr<<"SOAP: update: missing file id"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA;
    return SOAP_OK;
  };
  if(!(file->id)) file->id=(char*)(it->current_name);
  if(file->id) cerr<<"SOAP: update: file id: "<<file->id<<endl;
  if(file->size) cerr<<"SOAP: update: file size: "<<*(file->size)<<endl;
  if(file->checksum) cerr<<"SOAP: update: checksum: "<<(file->checksum)<<endl;
  if(file->acl) cerr<<"SOAP: update: acl: "<<(file->acl)<<endl;
  if(!(it->current_file)) {
    cerr<<"SOAP: update: there is no file to update"<<endl;
    r.error_code=ERR_INTERNAL_ERROR;
    return SOAP_OK;
  };
  if(file->acl) {
    cerr<<"SOAP: update: update does not accept ACL yet"<<endl;
    r.error_code=ERR_EXCESSIVE_INPUT_DATA;
    return SOAP_OK;
  };
  bool updated = false;
  if(file->size) {
    if(it->current_file->size() != *(file->size)) {
      cerr<<"SOAP: update: file size differs"<<endl;
      r.error_code=ERR_WRONG_INPUT_DATA;
      return SOAP_OK;
    };
  };
  if(file->checksum) {
    if(it->current_file->checksum_available()) {
      if(strcmp(file->checksum,it->current_file->checksum().c_str()) != 0) {
        cerr<<"SOAP: update: file checksum differs"<<endl;
        r.error_code=ERR_WRONG_INPUT_DATA;
        return SOAP_OK;
      };
    } else {
      it->current_file->checksum(file->checksum);
      updated=true;
    };
  };
// TODO - created

  SEFile& f = *(it->current_file);
  if((f.complete()) &&                            // metadata complete
     (f.state_reg() == REG_STATE_LOCAL)) {        // not registered yet
          // register new file
    if(f.state(FILE_STATE_UNCHANGED,REG_STATE_REGISTERING)) {
      if(it->files()->NS()->Register(f) != 0) {
        if(it->handle->register_retry()) {
          cerr<<"Failed to register (will retry)"<<endl;
        } else {
          cerr<<"Failed to register (what to do ?)"<<endl;
        };
        f.state(FILE_STATE_UNCHANGED,REG_STATE_LOCAL);
      } else {
        f.state(FILE_STATE_UNCHANGED,REG_STATE_ANNOUNCED);
      };
    } else {
      if(it->handle->register_retry()) {
        cerr<<"Failed to set REGISTERING state (will retry)"<<endl;
      } else {
        cerr<<"Failed to set REGISTERING state (what to do ?)"<<endl;
      };
    };
  };
  return SOAP_OK;
}

int ns__add(struct soap* sp,ns__fileinfo *file, struct ns__addResponse &r) {

  // initialize response
  r.error_code=0; r.sub_error_code=0;
  r.file.id=NULL;
  r.file.size=NULL;
  r.file.checksum=NULL;
  r.file.acl=NULL;
  r.file.url=NULL;
  r.file.state=NULL;
  
  // check if all necessary attributes provided
  if(!file) {
    cerr<<"SOAP: add: missing information about file"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    return SOAP_OK;
  };
  if(file->id) cerr<<"SOAP: add: file id: "<<file->id<<endl;
  if(file->size) cerr<<"SOAP: add: file size: "<<*(file->size)<<endl;
  if(file->checksum) cerr<<"SOAP: add: checksum: "<<(file->checksum)<<endl;
  if(file->acl) cerr<<"SOAP: add: acl: "<<(file->acl)<<endl;

  r.file.id=file->id;
/*
//  if((!file->id) || (!file->size) || (!file->checksum) || (!file->acl)) {
// !!!!!!!!!!!!!!!!!!!!
  if((!file->id) || (!file->size) || (!file->acl)) {
    cerr<<"SOAP: add: file information is not complete"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    return SOAP_OK;
  };
*/
  if(file->url) {
    cerr<<"SOAP: add: add does not need contact url in input parameters"<<endl;
    r.error_code=ERR_EXCESSIVE_INPUT_DATA; 
    return SOAP_OK;
  };
  HTTP_SE* it = (HTTP_SE*)(sp->user);
  if(it->current_name && (*it->current_name)) {
    cerr<<"SOAP: add: add requires only service in path"<<endl;
    r.error_code=ERR_EXCESSIVE_INPUT_DATA; 
    return SOAP_OK;
  };
  
  // attributes for response
  r.file.id=file->id;
  r.file.size=file->size;
  r.file.checksum=file->checksum;
  r.file.acl=file->acl;

  if(!file->id) {
    cerr<<"SOAP: add: missing file id (lfn)"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    return SOAP_OK;
  };
  if(!file->size) {
    cerr<<"SOAP: add: missing file size (needed for allocation)"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    return SOAP_OK;
  };
  if(!file->acl) {
    cerr<<"SOAP: add: missing ACL (can't access file without ACL)"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    return SOAP_OK;
  };
  // create new file
  SEAttributes attr(file->id,*(file->size));
  if(file->checksum) attr.checksum(string(file->checksum));
  if((!attr.complete()) && (it->handle->register_immediately())) {
    cerr<<"SOAP: add: missing file information needed for registration"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    return SOAP_OK;
  };
  SEFile* f = it->new_file(attr);
  if(f==NULL) {
    cerr<<"File not created: "<<file->id<<endl;
    r.error_code=ERR_CREATING_NEW_FILE;  // TODO: check for reason of failure
    return SOAP_OK;
  };
  string url_ = it->url;
  add_path_to_url(url_,file->id,false);
  cerr<<"SOAP: add: contact url: "<<url_<<endl;
  if((r.file.url=(char*)soap_strdup_l(sp,url_.c_str(),url_.length()+1)) == NULL) {
    cerr<<"SOAP: add: soap_strdup_l failed"<<endl;
    f->destroy();
    r.error_code=ERR_INTERNAL_ERROR; 
    return SOAP_OK;
  }; 
  // register new file
  if(it->handle->register_immediately()) { 
    if(!f->state(FILE_STATE_UNCHANGED,REG_STATE_REGISTERING)) {
      if(it->handle->register_retry()) { 
        cerr<<"SOAP: add: failed to set REG_STATE_REGISTERING (will retry)"<<endl;
        f->state(FILE_STATE_UNCHANGED,REG_STATE_LOCAL);
      } else {
        cerr<<"SOAP: add: failed to set REG_STATE_REGISTERING"<<endl;
        f->destroy();
        r.error_code=ERR_INTERNAL_ERROR; 
        return SOAP_OK;
      };
    } else {
      if(it->files()->NS()->Register(*f) != 0) {
        if(it->handle->register_retry()) { 
          cerr<<"SOAP: add: failed to register (will retry)"<<endl;
          f->state(FILE_STATE_UNCHANGED,REG_STATE_LOCAL);
        } else {
          cerr<<"SOAP: add: failed to register"<<endl;
          f->destroy();
          r.error_code=ERR_INTERNAL_ERROR; 
          return SOAP_OK;
        };
      } else {
        f->state(FILE_STATE_UNCHANGED,REG_STATE_ANNOUNCED);
      };
    };
  };
  if(!f->state(FILE_STATE_COLLECTING,REG_STATE_UNCHANGED)) {
    cerr<<"SOAP: add: failed to set FILE_STATE_COLLECTING"<<endl;
    it->files()->NS()->Unregister(*f);  // TODO: if failed set state to repeat later
    f->destroy();
    r.error_code=ERR_INTERNAL_ERROR; 
    return SOAP_OK;
  };
  return SOAP_OK;
}

/* report information about file */
int ns__info(soap *sp, ns__infoResponse &r) {
  // initialize response
  r.error_code=0; r.sub_error_code=0;
  r.file.id=NULL;
  r.file.size=NULL;
  r.file.checksum=NULL;
  r.file.acl=NULL;
  r.file.url=NULL;
  r.file.state=NULL;
  HTTP_SE* it = (HTTP_SE*)(sp->user);
  SEUserDescription u_desc;
  u_desc.subject=it->c->identity_subject();
  int acl_flags = it->current_file->check_acl(u_desc);
  if(acl_flags & FILE_ACC_LIST) {
    r.file.id=soap_strdup_l(sp,it->current_name);
    r.file.size=(uint64_t*)soap_malloc(sp,sizeof(uint64_t)); if(r.file.size) (*r.file.size)=it->current_file->size();
    r.file.checksum=soap_strdup_l(sp,it->current_file->checksum().c_str());
    string acl;
    if(it->current_file->read_acl(u_desc,acl) == 0) {
      r.file.acl=soap_strdup_l(sp,acl.c_str(),acl.length());
    };
    string url_ = it->url + "/" + it->current_file->id();
    r.file.url=(char*)soap_strdup_l(sp,url_.c_str(),url_.length()+1);
    r.file.state=(ns__filestate*)soap_malloc(sp,sizeof(ns__filestate)); if(r.file.state) (*r.file.state)=(ns__filestate)(it->current_file->state());
  } else {
    r.error_code=ERR_INTERNAL_ERROR;
  };
  return SOAP_OK;
}

/* report/modify ACL */
int ns__acl(soap *sp, char *acl, ns__aclResponse &r) {
  int res = -1;
  r.error_code=SUCCESS;
  r.sub_error_code=0;
  r.acl=NULL;
  HTTP_SE* it = (HTTP_SE*)(sp->user);
  SEUserDescription u_desc;
  u_desc.subject=it->c->identity_subject();
  int acl_flags = it->current_file->check_acl(u_desc);
  if(acl) {
    if(acl_flags & FILE_ACC_WRITE_META) {
      res=it->current_file->write_acl(u_desc,acl);
    };
  } else {
    if(acl_flags & FILE_ACC_READ_META) {
      string acl_;
      res=it->current_file->read_acl(u_desc,acl_);
      if(res==0) {
        if((r.acl=(char*)soap_malloc(sp,acl_.length()+1)) == NULL) {
          r.error_code=ERR_INTERNAL_ERROR;
          return SOAP_OK;
        };
        strcpy(r.acl,acl_.c_str());
      };
    };
  };
  if(res!=-1) {
    r.error_code=ERR_INTERNAL_ERROR;
  };
  return SOAP_OK;
}

int ns__del(soap *sp, ns__delResponse &r) {
  // initialize response
  r.error_code=0; r.sub_error_code=0;
  HTTP_SE* it = (HTTP_SE*)(sp->user);
  SEUserDescription u_desc;
  u_desc.subject=it->c->identity_subject();
  int acl_flags = it->current_file->check_acl(u_desc);
  if(!(acl_flags & FILE_ACC_DELETE)) {
    r.error_code=ERR_INTERNAL_ERROR;
    return SOAP_OK;
  };
  if(it->files()->NS()->Unregister(*(it->current_file)) != 0) { // TODO: for lazy synch repeat later
    r.error_code=ERR_INTERNAL_ERROR;
    return SOAP_OK;
  };
  if(it->files()->remove(*(it->current_file)) != 0) {
    r.error_code=ERR_INTERNAL_ERROR;
    return SOAP_OK;
  };
  return SOAP_OK;
}

