#include <iostream>
#include <unistd.h>
#include <string>
#include <vector>
#include <sys/time.h>
#include <sys/wait.h>
using namespace std;

static int number_testcases = 85;

void print_usage( string progname );
void print_cases( );

void executable_args( vector<string> &components );
void arguments_args( vector<string> &components );
void executables_args( vector<string> &components );
void inputfiles_args( vector<string> &components );
void outputfiles_args( vector<string> &components );
void runtimeenvironment_args( vector<string> &components );
string convert_int_to_string( unsigned int number );

int main( int argv, char* argc[] )
{
  string progname = argc[0];
  unsigned int pos = progname.rfind('/');
  if (pos != string::npos)
    progname = progname.substr (pos + 1);

  vector<string> jobs;
  vector<string> ngsubargs;

  bool submitseparatly = false;

  jobs.clear();
  ngsubargs.clear();

  // initial parsing of specified arguments
  for(int i=1; i<argv; i++) {

    if( (strcmp(argc[i],"-v")==0)||(strcmp(argc[i],"--version")==0) ) {
      cout << progname << ": version " << VERSION << endl;
      exit(0);
    }
    
    if( (strcmp(argc[i],"-h")==0)||(strcmp(argc[i],"--h")==0)||(strcmp(argc[i],"--help")==0) ) {
      print_usage( progname );
      exit(0);
    }

    if( (strcmp(argc[i],"-list-cases")==0) ) {
      print_cases( );
      exit(0);
    }

    if( (strcmp(argc[i],"-s")==0) ) {
      submitseparatly = true;
      continue;
    }

    if( argc[i][0]=='-' ) {
      ngsubargs.push_back(argc[i]);
      if(i<argv-1) {
	ngsubargs.push_back(argc[i+1]);
	i++;
      }
    } else {
      jobs.push_back( argc[i] );
    }				
  }

  // cout << "ngsubargs: " << ngsubargs << endl;
  // find all jobids
  vector<int> jobids;

  for(unsigned int i=0; i<jobs.size(); i++) {
    if( jobs[i].find_first_not_of("0123456789-:") != string::npos ) {
      cout << "Do not understand jobid: " << jobs[i] << endl;
      exit(1);
    }

    if( (jobs[i].find_first_of(":") != string::npos) && (jobs[i].find_first_of("-") != string::npos) ) {
      cout << "Cannot specify both : and - in a jobid: " << jobs[i] << endl;
      exit(1);
    }

    if( jobs[i].find_first_of(":") != string::npos ) {
      // there is a colon in the jobid
      
      unsigned int pos = 0;
      if( (pos = jobs[i].find_first_of(":")) != jobs[i].find_last_of(":") ) {
	cout << "Can only specify one : in a jobid: " << jobs[i] << endl;
	exit(1);
      }
      // there is only one colon in the jobid

      // find the two numbers to the left and right of the colon
      int start = atoi( jobs[i].substr( 0,pos ).c_str() );
      int number = atoi( jobs[i].substr( pos+1 ).c_str() );

      // build jobids
      for(int id=start; id<start+number; id++)
	if( (id>=0) && (id<=number_testcases) ) {
	  jobids.push_back( id );
	} else {
	  cout << "Ignoring jobid: " << id << endl;
	}
    } else if( jobs[i].find_first_of("-") != string::npos ) {
      // there is a minus in the jobid

      unsigned int pos = 0;
      if( (pos = jobs[i].find_first_of("-")) != jobs[i].find_last_of("-") ) {
        cout << "Can only specify one - in a jobid: " << jobs[i] << endl;
        exit(1);
      }
      // there is only one minus in the jobid

      // find the two numbers to the left and right of the minus
      int start = atoi( jobs[i].substr( 0,pos ).c_str() );
      int end   = atoi( jobs[i].substr( pos+1 ).c_str() );

      // build jobids
      for(int id=start; id<=end; id++)
	if( (id>=0) && (id<=number_testcases) ) {
	  jobids.push_back( id );
	} else {
	  cout << "Ignoring jobid: " << id << endl;
	}
    } else {
      // there are no colons nor minuses in the jobid
      int id = atoi( jobs[i].c_str() );
      if( (id>=0) && (id<=number_testcases) ) {
	jobids.push_back( id );
      } else {
	cout << "Ignoring jobid: " << id << endl;
      }
    }

  }

  if(jobids.size()==0) {
    cout << "No testjob specified. Submitting default test-job 0." << endl;
    jobids.push_back( 0 );
  }

  // sort the jobids
  cout << "Submitting test-job numbers: ";

  for( unsigned int i=0; i<jobids.size(); i++)
    cout << jobids[i] << " "; 
  cout << endl;

  // constructing test-jobs
  vector<string> executable, arguments, executables, inputfiles, runtimeenvironments, outputfiles;
  executable_args( executable );
  arguments_args( arguments );
  executables_args( executables );
  inputfiles_args( inputfiles );
  outputfiles_args( outputfiles );
  runtimeenvironment_args( runtimeenvironments );

  // now construct the job-xrsl
  const unsigned int numberofjobs = jobids.size();
  string jobxrsl[numberofjobs];

  for( unsigned int i=0; i<jobids.size(); i++) {
    
    jobxrsl[i] += "&";  // initial &

    // now build test-jobs
    if( jobids[i]==0 ) {
      jobxrsl[i] += executable[7];
      jobxrsl[i] += "(inputfiles=";

      jobxrsl[i] = jobxrsl[i] + "(\"run.sh\" \"http://www.nordugrid.org/data/run.sh\")";
      jobxrsl[i] = jobxrsl[i] + "(\"Makefile\" \"gsiftp://dc3.uio.no/dc1/test/ngtest/Makefile\")";
      jobxrsl[i] = jobxrsl[i] + "(\"prime.cpp\" \"ftp://ftp.nordugrid.org/applications/test/prime.cpp\"))";

      jobxrsl[i] += "(stderr=\"primenumbers\")";

      jobxrsl[i] += "(outputfiles=";
      jobxrsl[i] = jobxrsl[i] + "(\"primenumbers\" \"primenumbers\"))";
    }

    if( jobids[i]==1 ) {
      jobxrsl[i] += executable[0];
      jobxrsl[i] += arguments[0];
    }

    if( jobids[i]==2 ) {
      jobxrsl[i] += executable[1];
      jobxrsl[i] += arguments[0];
    }

    if( jobids[i]==3 ) {
      jobxrsl[i] += executable[2];
      jobxrsl[i] += arguments[0];
      jobxrsl[i] += runtimeenvironments[0];
    }      

    if( jobids[i]==4 ) {
      jobxrsl[i] += executable[0];
      jobxrsl[i] += arguments[1];
    }      

    if( jobids[i]==5 ) {
      jobxrsl[i] += executable[0];
      jobxrsl[i] += arguments[2];
    }      

    if( jobids[i]==6 ) {
      jobxrsl[i] += executable[4];
      jobxrsl[i] += arguments[6];
      jobxrsl[i] += runtimeenvironments[0];
    }

    if( jobids[i]==7 ) {
      jobxrsl[i] += executable[5];
      jobxrsl[i] += runtimeenvironments[0];
      jobxrsl[i] += runtimeenvironments[1];
    }

    if( jobids[i]==8 ) {
      jobxrsl[i] += executable[4];
      jobxrsl[i] += arguments[3];
      jobxrsl[i] += executables[0];
      jobxrsl[i] = jobxrsl[i] + "(inputfiles=(\"echo.sh\" " +inputfiles[0] + "))";
    }      

    if( (jobids[i]>8) && (jobids[i]<61) ) {
      
      if( (jobids[i]-9) % 13 != 12) {
	
	int type = ( (jobids[i]-9) % 13 ) / 3;
	int number = ( (jobids[i]-9) % 13 ) % 3;

	if( (number==0)||(number==1) ) {
	  jobxrsl[i] += executable[0];
	  jobxrsl[i] += arguments[0];
	} else {
	  jobxrsl[i] += executable[3];
	  jobxrsl[i] += arguments[5];
	}

	jobxrsl[i] += "(inputfiles=";

	if(number==2) number=8;
	
	for(int j=0; j<=number; j++) {
	  jobxrsl[i] = jobxrsl[i] + "(" + "\"downloaded" + (char)(48+j) + "\" ";
	  jobxrsl[i] = jobxrsl[i] + inputfiles[3*type+(j % 3)] + ")";
	}

	if(number==8) jobxrsl[i] += "(\"md5sum.sh\" " + inputfiles[15] + ")";
	jobxrsl[i] += ")";
      }      
      
      if( (jobids[i]-9) % 13 == 12 ) {
	jobxrsl[i] += executable[3];
	jobxrsl[i] += arguments[5];

	jobxrsl[i] += "(inputfiles=";
	for(int j=0; j<=7; j++) {
	  jobxrsl[i] = jobxrsl[i] + "(" + "\"downloaded" + (char)(48+j) + "\" ";
	  jobxrsl[i] = jobxrsl[i] + inputfiles[3*(j / 2)+(j % 2)] + ")";
	}

	jobxrsl[i] += "(\"md5sum.sh\" " + inputfiles[15] + ")";
	jobxrsl[i] += ")";

      }
      
      if( (jobids[i]-9) >= 26 )
	jobxrsl[i] += "(ftpThreads=10)";
      if( ((jobids[i]-9)/13==1) || ((jobids[i]-9)/13==3) ) {
	unsigned int j = 0;
	while( (j = jobxrsl[i].find( "yes" )) != string::npos )
	  jobxrsl[i].replace( j, 3, "no");
      }
    }

    if( (jobids[i]>=61) && (jobids[i]<=66) ) {
      jobxrsl[i] += executable[3];
      jobxrsl[i] += arguments[5];
	
      int number = (jobids[i]-61) % 3;
      if(number==2) number=8;

      jobxrsl[i] += "(inputfiles=";
      for(int j=0; j<=number; j++) {
	jobxrsl[i] = jobxrsl[i] + "(" + "\"downloaded" + (char)(48+j) + "\" ";
	jobxrsl[i] = jobxrsl[i] + inputfiles[12+(j % 3)] + ")"; 
      }
      jobxrsl[i] += "(\"md5sum.sh\" " + inputfiles[15] + ")";
      jobxrsl[i] += ")";

      if( jobids[i]>=64 )
	jobxrsl[i] += "(ftpThreads=10)";

    }

    if( (jobids[i]==67)||(jobids[i]==68) ) {
      jobxrsl[i] += executable[3];
      jobxrsl[i] += arguments[5];

      jobxrsl[i] += "(inputfiles=";
      for(int j=0; j<=9; j++) {
	jobxrsl[i] = jobxrsl[i] + "(" + " \"downloaded" + (char)(48+j) + "\" ";
	jobxrsl[i] = jobxrsl[i] + inputfiles[3*(j / 2)+(j % 2)] + ")";
      }
      jobxrsl[i] += "(\"md5sum.sh\" " + inputfiles[15] + ")";
      jobxrsl[i] += ")";

      if( jobids[i]==68 )
	jobxrsl[i] += "(ftpThreads=10)";
    }

    if( (jobids[i]>=69)&&(jobids[i]<=71) ) {
      jobxrsl[i] += executable[0];
      jobxrsl[i] += arguments[0];
      
      if(jobids[i]==69) jobxrsl[i] += "(cluster=\"lscf\")";
      if(jobids[i]==70) jobxrsl[i] += "(|(cluster=\"grid.tsl\")(cluster=\"grid.quark\"))";
      if(jobids[i]==71) jobxrsl[i] += "(cluster!=\"lscf\")";
    }
    
    if( jobids[i]==72 ) {
      jobxrsl[i] += executable[0];
      jobxrsl[i] += arguments[0];
      
      jobxrsl[i] += "(cluster=\"grid.nbi\")";
      jobxrsl[i] += "(queue=\"short\")";
    }

    if( (jobids[i]>=73)&&(jobids[i]<=75) ) {
      jobxrsl[i] += executable[0];
      jobxrsl[i] += arguments[0];
      
      if(jobids[i]==73) jobxrsl[i] += "(middleware=\"nordugrid-0.3.24\")";
      if(jobids[i]==74) jobxrsl[i] += "(middleware>=\"nordugrid-0.3.24\")";
      if(jobids[i]==75) jobxrsl[i] += "(middleware<=\"nordugrid-0.3.24\")";
    }

    if( jobids[i]==76 ) {
      jobxrsl[i] += executable[6];
      jobxrsl[i] += arguments[0];

      jobxrsl[i] += "(environment=(\"ECHOBIN\" \"/bin\"))";
    }
    
    if( jobids[i]==77 ) {
      jobxrsl[i] += "(rsl_substitution=(\"ECHOBIN\" \"/bin\"))";
      jobxrsl[i] += executable[6];
      jobxrsl[i] += arguments[0];

    }
    
    if( jobids[i]==78 ) {
      jobxrsl[i] += executable[0];
      jobxrsl[i] += "(rsl_substitution=(\"ECHODIR\" \".\"))";
      jobxrsl[i] += arguments[6];
    }

    if( jobids[i]==79 ) { 
      jobxrsl[i] += executable[0];
      jobxrsl[i] += arguments[0];
      jobxrsl[i] += "(cluster=lscf)(queue=gridshort)(CpuTime=\"2 hours\")";
    }

    if( (jobids[i]>=80)&&(jobids[i]<=85) ) {
      struct timeval time;
      struct timezone zone;
      int result = gettimeofday( &time, &zone );
      if ( result != 0 ) {
	cout << "Could not gettimeofday! Exiting." << endl;
	return 1;
      }

      unsigned int seed = time.tv_usec;
      srandom( seed );

      jobxrsl[i] += executable[0];
      
      string argument_string = "The following files have been uploaded:";

      int number = (jobids[i]-80) % 3;
      int type = (jobids[i]-80) / 3;
      if(number==2) number = 8;

      jobxrsl[i] += "(inputfiles=";
      for(int j=0; j<=number; j++) {
	jobxrsl[i] = jobxrsl[i] + "(" + "\"downloaded" + (char)(48+j) + "\" ";
	jobxrsl[i] = jobxrsl[i] + inputfiles[6+(j % 3)] + ")"; 
      }
      jobxrsl[i] += ")";

      jobxrsl[i] += "(outputfiles=";
      for(int j=0; j<=number; j++) {
	unsigned int randomnumber = (unsigned int)random();
	string randomnumber_str = convert_int_to_string( randomnumber );

	jobxrsl[i] = jobxrsl[i] + "(" + "\"downloaded" + (char)(48+j) + "\" ";
	jobxrsl[i] = jobxrsl[i] + "\"" + outputfiles[type] + "uploaded" + randomnumber_str + "\")";
	argument_string = argument_string + " " + outputfiles[type] + "uploaded" + randomnumber_str;
      }
      jobxrsl[i] += ")";
      argument_string += ". They can be removed with ngremove.";

      jobxrsl[i] = jobxrsl[i] + "(arguments=\"" + argument_string + "\")";

    }

    string jobid_str = convert_int_to_string( jobids[i] );
    
    jobxrsl[i] = jobxrsl[i] + "(jobName=\"ngtest-job-" + jobid_str + "\")";
    jobxrsl[i] = jobxrsl[i] + "(stdout=\"stdout\")";

    if( jobids[i]!=0 ) {
      jobxrsl[i] = jobxrsl[i] + "(stderr=\"stderr\")";
      jobxrsl[i] = jobxrsl[i] + "(gmlog=\"gmlog\")";
    }

    jobxrsl[i] = jobxrsl[i] + "(CPUTime=100)";

  }

  //Now construct the arguments for the ngsub-command
  int numargs = ngsubargs.size();
  char* ngsub_argv[numargs+3];

  ngsub_argv[0] = "ngsub";
  for(int i=0; i<numargs; i++) {
    ngsub_argv[i+1] = (char *)ngsubargs[i].c_str();
  }

  ngsub_argv[numargs+2] = 0;

  //Submit xrsl
  string xrsl;
  if( submitseparatly == false ) {
    
    if( numberofjobs>1 ) xrsl = "+";
    
    for( unsigned int i=0; i<numberofjobs; i++ ) {
      
      if( numberofjobs>1 ) xrsl += "(";
      xrsl += jobxrsl[i];
      if( numberofjobs>1 ) xrsl += ")";
    
    }
    
    cout << "Submitting xrsl: " << xrsl << endl;

    ngsub_argv[numargs+1] = (char *)xrsl.c_str();
    execvp("ngsub", ngsub_argv);
  }

  if( submitseparatly == true ) {
    for( unsigned int i=0; i<numberofjobs; i++ ) {

      ngsub_argv[numargs+1] = (char *)jobxrsl[i].c_str();

      // now fork before submitting
      pid_t pid;
      pid = fork();

      switch( pid ) {

      case -1:
	cout << "Fork failed. Exiting." << endl;
	exit(1);

      case 0:
	cout << "Submitting xrsl: " << jobxrsl[i] << endl;
	execvp("ngsub", ngsub_argv);

      default:
	int child_status;
	pid_t child_pid;

	child_pid = wait( &child_status );
	if( WIFEXITED(child_status)==0 ) {
	  cout << "Job " << jobids[i] << " not submitted! Exiting" << endl;
	  exit(1);
	}
      }
    }

    cout << "All jobs submitted successfully." << endl;
  }

  return 0;
}

string convert_int_to_string( unsigned int number )
{
  string number_str = "";
    
  if(number == 0) number_str = "0";
  while( number > 0) {
    number_str = (char)(48+(number % 10)) + number_str;
    number = number/10;
  }
  
  return number_str;
}

void executable_args( vector<string> &components )
{
  string begin = "(executable=";
  string end   = ")";

  components.push_back( begin + "\"/bin/echo\"" + end );
  components.push_back( begin + "\"echo.sh\""   + end );
  components.push_back( begin + "\"$(ECHOBIN)/echo.sh\"" + end );
  components.push_back( begin + "\"md5sum.sh\"" + end );
  components.push_back( begin + "\"/bin/ls\""   + end );
  components.push_back( begin + "\"/usr/bin/printenv\"" + end );
  components.push_back( begin + "$(ECHOBIN)/echo" + end );
  components.push_back( begin + "\"run.sh\"" + end );
}

void arguments_args( vector<string> &components )
{
  string begin = "(arguments=";
  string end   = ")";

  components.push_back( begin + "\"hello, grid\"" + end );
  components.push_back( begin + "1 2 3" + end );
  components.push_back( begin + "1 2 3 4 5 6 7 8 9 0" + end );
  components.push_back( begin + "-l" + end );
  components.push_back( begin + "\"echo.sh\"" + end );
  components.push_back( begin + "downloaded[0-9]" + end );
  components.push_back( begin + "$(ECHODIR)/echo.sh" + end );
  components.push_back( begin + "\"hello, testsuite\"" + end );
}

void executables_args( vector<string> &components )
{
  string begin = "(executables=\"";
  string end   = "\")";

  components.push_back( begin + "echo.sh" + end );
  components.push_back( begin + "cat.sh" + end );
}

void inputfiles_args( vector<string> &components )
{
  components.push_back( "\"http://www.nordugrid.org;cache=yes/data/echo.sh\"" );
  components.push_back( "\"http://www.nordugrid.org;cache=yes/data/somefile\"" );
  components.push_back( "\"http://www.nordugrid.org;cache=yes/data/somehugefile\"" );

  components.push_back( "\"gsiftp://dc3.uio.no;cache=yes/dc1/test/ngtest/echo.sh\"" );
  components.push_back( "\"gsiftp://dc3.uio.no;cache=yes/dc1/test/ngtest/somefile\"" );
  components.push_back( "\"gsiftp://dc3.uio.no;cache=yes/dc1/test/ngtest/somehugefile\"" );

  components.push_back( "\"ftp://ftp.nordugrid.org;cache=yes/applications/test/echo.sh\"" );
  components.push_back( "\"ftp://ftp.nordugrid.org;cache=yes/applications/test/somefile\"" );
  components.push_back( "\"ftp://ftp.nordugrid.org;cache=yes/applications/test/somehugefile\"" );

  components.push_back( "\"rc://;cache=yes@grid.uio.no/lc=TestCollection,rc=NorduGrid,dc=nordugrid,dc=org/ngtest/echo.sh\"" );
  components.push_back( "\"rc://;cache=yes@grid.uio.no/lc=TestCollection,rc=NorduGrid,dc=nordugrid,dc=org/ngtest/somefile\"" );
  components.push_back( "\"rc://;cache=yes@grid.uio.no/lc=TestCollection,rc=NorduGrid,dc=nordugrid,dc=org/ngtest/somehugefile\"" );

  components.push_back( "\"echo.sh\"" );
  components.push_back( "\"somefile\"" );
  components.push_back( "\"somehugefile\"" );

  // misc
  components.push_back( "\"ftp://ftp.nordugrid.org;cache=yes/applications/test/md5sum.sh\"" );
  components.push_back( "\"ftp://ftp.nordugrid.org;cache=yes/applications/test/ECHOENV\"");
  components.push_back( "\"ftp://ftp.nordugrid.org;cache=yes/applications/test/MD5SUMENV\"");

}

void outputfiles_args( vector<string> &components )
{
  components.push_back( "gsiftp://dc3.uio.no/dc1/test/ngtest/uploaded/" );
  components.push_back( "rc://dc3.uio.no@grid.uio.no/lc=TestCollection,rc=NorduGrid,dc=nordugrid,dc=org/ngtest/uploaded/" );
}

void runtimeenvironment_args( vector<string> &components )
{
  string begin = "(runTimeEnvironment=\"";
  string end   = "\")";

  components.push_back( begin + "ECHOENV" + end );
  components.push_back( begin + "MD5SUMENV" + end );
}

void print_usage( string progname )
{
  cout << "Usage: " << progname << " [jobids] [xrsl-options ...]" << endl;
  cout << endl;
  cout << "Jobids is a list of job-partitions which can take the form:" << endl;
  cout << endl;
  cout << "x    means: x" << endl;
  cout << "x:n  means: x, x+1,..., x+(n-1)" << endl;
  cout << "x-y  means: x,...,y" << endl;
  cout << endl;
  cout << "i.e. 4 5:3 14-17 gets expanded to:" << endl;
  cout << endl;
  cout << "4 5 6 7 14 15 16 17" << endl;
  cout << endl;
  cout << "The xrsl-options are passed on directly to ngsub." << endl;
  cout << endl;
  cout << "To see the full list of test-cases, run ngtest -list-cases" << endl;
}

void print_cases( )
{
  cout << "The following test-jobs are included in the program at the moment:" << endl;
  cout << endl;
  cout << "#0 : " << endl;
  cout << "#1 : basic executable xrsl (hello grid)" << endl;
  cout << "#2 : UI uploads executable                        (requires echo.sh locally)" << endl;
  cout << "#3 : executable specified via runtime-environment (requires ECHOENV on cluster)" << endl;
  cout << "#4 : tests  3 arguments" << endl;
  cout << "#5 : tests 10 arguments" << endl;
  cout << "#6 : tests runtime-environment in arguments-attr  (requires ECHOENV on cluster)" << endl;
  cout << "#7 : tests 2 runtime-environments                 (requires ECHOENV, MD5SUMENV)" << endl;
  cout << "#8 : test executables attribute                   (requires echo.sh locally)" << endl;
  cout << endl;
  cout << "#9 : tests downloading 1 inputfile via http" << endl;
  cout << "#10: tests downloading 2 inputfiles via http" << endl;
  cout << "#11: tests downloading 9 inputfiles via http" << endl;
  cout << "#12: tests downloading 1 inputfiles via gsiftp" << endl;
  cout << "#13: tests downloading 2 inputfiles via gsiftp" << endl;
  cout << "#14: tests downloading 9 inputfiles via gsiftp" << endl;
  cout << "#15: tests downloading 1 inputfile via ftp" << endl;
  cout << "#16: tests downloading 2 inputfiles via ftp" << endl;
  cout << "#17: tests downloading 9 inputfiles via ftp" << endl;
  cout << "#18: tests downloading 1 inputfiles via rc (and gsiftp)" << endl;
  cout << "#19: tests downloading 2 inputfiles via rc (and gsiftp)" << endl;
  cout << "#20: tests downloading 9 inputfiles via rc (and gsiftp)" << endl;
  cout << "#21: tests downloading 8 files, 2 of each via http, ftp, gsiftp and rc" << endl;
  cout << endl;
  cout << "#22-#34: as above with cache=no" << endl;
  cout << "#35-#47: as above with ftpThreads=10" << endl;
  cout << "#48-#60: as above with cache=no and ftpThreads=10" << endl;
  cout << endl;
  cout << "#61: UI uploads 1 file      (requires echo.sh locally)" << endl;
  cout << "#62: UI uploads 2 files     (requires echo.sh, somefile locally)" << endl;
  cout << "#63: UI uploads 9 files     (requires echo.sh, somefile, somehugefile locally)" << endl;
  cout << endl;
  cout << "#64-#66: as above with ftpThreads=10" << endl;
  cout << endl;
  cout << "#67: combination of #21 and #62 -- 10 files processed" << endl;
  cout << "#68: as #67 with ftpThreads=10" << endl;
  cout << endl;
  cout << "#69: tests cluster-attribute" << endl;
  cout << "#70: tests cluster-attribute and |" << endl;
  cout << "#71: tests cluster-attribute and !=" << endl;
  cout << "#72: tests cluster- and queue attribute" << endl;
  cout << endl;
  cout << "#73: tests middleware-attribute (=)" << endl;
  cout << "#74: tests middleware-attribute (>=)" << endl;
  cout << "#75: tests middleware-attribute (<=)" << endl;
  cout << endl;
  cout << "#76: tests environment-attribute" << endl;
  cout << "#77: tests rsl_substitution in executable-attribute" << endl;
  cout << "#78: tests rsl_substitution in arguments-attribute" << endl;
  cout << "#79: tests brokering and CpuTime with \"xx hours\"  (see ng-bug-63)" << endl;

  cout << "#80: uploading 1 file to gsiftp-server" << endl;
  cout << "#81: uploading 2 file to gsiftp-server" << endl;
  cout << "#82: uploading 9 file to gsiftp-server" << endl;
  cout << "#83: uploading 1 file to replica-catalog" << endl;
  cout << "#84: uploading 2 file to replica-catalog" << endl;
  cout << "#85: uploading 9 file to replica-catalog" << endl;

}





